(ns com.aurellem.gb.util
  (:use (com.aurellem.gb gb-driver vbm))
  (:import [com.aurellem.gb.gb_driver SaveState]))


(defn A [state]
  (bit-shift-right (bit-and 0x0000FF00 (AF state)) 8))

(defn B [state]
  (bit-shift-right (bit-and 0x0000FF00 (BC state)) 8))

(defn D [state]
  (bit-shift-right (bit-and 0x0000FF00 (DE state)) 8))

(defn H [state]
  (bit-shift-right (bit-and 0x0000FF00 (HL state)) 8))

(defn C [state]
  (bit-and 0xFF (BC state)))
(defn F [state]
  (bit-and 0xFF (AF state)))
(defn E [state]
  (bit-and 0xFF (DE state)))
(defn L [state]
  (bit-and 0xFF (HL state)))

(defn binary-str [num]
  (format "%08d"
          (Integer/parseInt
           (Integer/toBinaryString num) 10)))

(defn view-register [state name reg-fn]
  (println (format "%s: %s" name
                   (binary-str (reg-fn state))))
  state)

(defn view-memory
  ([^SaveState state mem]
     (println (format "mem 0x%04X = %s" mem
                      (binary-str (aget (memory state) mem))))
     state)
  ([mem]
     (view-memory @current-state mem)))

(defn print-listing
  ([^SaveState state begin end]
     (dorun (map 
             (fn [opcode line]
               (println (format "0x%04X:  0x%02X" line opcode)))
             (subvec  (vec (memory state)) begin end)
             (range begin end)))
     state)
  ([begin end]
     (print-listing @current-state begin end)))

(defn print-pc
  ([^SaveState state]
     (println (format "PC: 0x%04X" (PC state)))
     state)
  ([] (print-pc @current-state))) 

(defn print-op
  ([^SaveState state]
     (println (format "OP: 0x%02X" (aget (memory state) (PC state))))
     state)
  ([] (print-op @current-state)))

(defn d-tick
  ([state]
  (-> state print-pc print-op tick)))

(defn print-interrupt
  [^SaveState state]
  (println (format "IE: %d" (IE state)))
  state)

(defn set-memory
  ([state location value]
     (set-state! state)
     (let [mem (memory state)]
       (aset mem location value)
       (write-memory! mem)
       (update-state)))
  ([location value]
     (set-memory @current-state location value)))

(defn set-memory-range
  ([state start values]
     (set-state! state)
     (let [mem (memory state)]
  
       (dorun (map (fn [index val]
                     (aset mem index val))
                   (range start
                          (+ start (count values))) values))
    (write-memory! mem)
    (update-state)))
  ([start values]
     (set-memory-range
      @current-state start values)))

(defn common-differences [& seqs]
  (let [backbone (range (count (first seqs)))]
    (filter
     (comp (partial apply distinct?) second)
     (zipmap backbone
             (apply (partial map list) seqs)))))

(defn mid-game []
  (read-state "mid-game"))



(defn disect-bytes-2
  "return a vector consiting of the last 16 bytes of the
   integer expressed as two 8 bit nimbers (inside an integer)
   in the form [high-bits low-bits."
  [num]
  [(bit-shift-right
    (bit-and num 0xFF00) 8)
   (bit-and num 0xFF)])

(defn disect-bytes-3
  "same as disect-bytes-2 except that it assumes the input is a
   24 bit number and returns [high-bits medium-bits low-bits]"
  [num]
  (vec
   (concat 
    [(bit-shift-right (bit-and num 0xFF0000) 16)]
    (disect-bytes-2 num))))

(defn glue-bytes
  "Given two or three 8-bit numbers inside 32-bit integers,
   combine them into the integer number that they together
   represent."
  ([h l]
     (+ l (bit-shift-left h 8)))

  ([h m l]
     (+ (glue-bytes m l)
        (bit-shift-left h 16))))


